#
#   CodeLogician Server
#
#   main.py
#

from termcolor import colored
import argparse, uvicorn, logging, sys, os, dotenv
from pathlib import Path

dotenv.load_dotenv(".env")
if 'IMANDRA_UNI_KEY' not in os.environ:
    print ("CodeLogician requires 'IMANDRA_UNI_KEY' to be set!")
    sys.exit(0)

log = logging.getLogger(__name__)
logging.getLogger("httpx").setLevel(logging.WARNING)
logging.getLogger("fsevents").setLevel(logging.WARNING)
#logging.getLogger("strategy.pyiml_strategy").setLevel(logging.WARNING)
#logging.getLogger("strategy.cl_worker").setLevel(logging.WARNING)
#logging.getLogger("strategy.metamodel").setLevel(logging.WARNING)
log.setLevel(logging.INFO)

from .cl_server import CLServer
from .oneshot import do_oneshot
from .config import ServerConfig
from .state import ServerState
from .endpoints import register_endpoints

ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.INFO) # Only show INFO and above
ch.setFormatter(logging.Formatter("%(asctime)s [%(processName)s: %(process)d] [%(threadName)s: %(thread)d] [%(levelname)s] %(name)s: %(message)s"))
log.addHandler(ch)


def do_intro():

    text = """


   █████████               █████          █████                          ███            ███                                       ████████ 
  ███░░░░░███             ░░███          ░░███                          ░░░            ░░░                                       ███░░░░███
 ███     ░░░   ██████   ███████   ██████  ░███         ██████   ███████ ████   ██████  ████   ██████   ████████      █████ █████░░░    ░███
░███          ███░░███ ███░░███  ███░░███ ░███        ███░░███ ███░░███░░███  ███░░███░░███  ░░░░░███ ░░███░░███    ░░███ ░░███    ███████ 
░███         ░███ ░███░███ ░███ ░███████  ░███       ░███ ░███░███ ░███ ░███ ░███ ░░░  ░███   ███████  ░███ ░███     ░███  ░███   ███░░░░  
░░███     ███░███ ░███░███ ░███ ░███░░░   ░███      █░███ ░███░███ ░███ ░███ ░███  ███ ░███  ███░░███  ░███ ░███     ░░███ ███   ███      █
 ░░█████████ ░░██████ ░░████████░░██████  ███████████░░██████ ░░███████ █████░░██████  █████░░████████ ████ █████     ░░█████   ░██████████
  ░░░░░░░░░   ░░░░░░   ░░░░░░░░  ░░░░░░  ░░░░░░░░░░░  ░░░░░░   ░░░░░███░░░░░  ░░░░░░  ░░░░░  ░░░░░░░░ ░░░░ ░░░░░       ░░░░░    ░░░░░░░░░░ 
                                                               ███ ░███                                                                    
                                                              ░░██████                                                                     
                                                               ░░░░░░                                                                      

                                                             
"""

    print (colored(text, "blue"))

def set_server_arguments(parser):
    """
    Set the server arguments to the command arguments parser.
    """
    parser.add_argument("-s", "--state", type=str, default=None, help="Server state file to use.")
    parser.add_argument("-d", "--dir", type=str, default=None, help="Target directory. This takes precedence over `strat_config`.")
    parser.add_argument("-c", "--clean", action="store_true", help="Start clean by disregarding any existing cache files.")
    parser.add_argument("--config", type=str, default="config/server_config.yaml", help="Server configuration YAML file.")
    parser.set_defaults(func=run_server)

def run_server(args):
    """ 
    Run the server
    """

    do_intro()

    try:
        config = ServerConfig.fromYAML(args.config)
    except Exception as e:
        print(f"Failed to load in server config: {str(e)}. Using defaults.")
        config = ServerConfig()


    if args.state:
        # We need to use the existing state
        abs_path = str(Path(args.state).resolve())

        if not os.path.exists(abs_path):
            log.error(f"Specified path for server config doesn't exist: [{abs_path}]. Using defaults.")
            state = ServerState(abs_path=abs_path)
        else:
            try:
                state = ServerState.fromFile(abs_path)
            except Exception as e:
                log.error(f"Failed to create server state from specified file: {abs_path}")
                raise Exception (f"Failed to read in server state: {str(e)}")

    else:
        # We're creating a new state 
        server_state_abs_path = os.path.join(os.getcwd(), '.cl_server')

        try:
            state = ServerState.fromFile(server_state_abs_path)
        except Exception as e:
            log.warning (f"Failed to create server state file: {e}")
            state = ServerState(abs_path=server_state_abs_path, strategy_paths=[], config=config)

        if args.dir:
            # We need to add a strategy directory to the state
            abs_path = str(Path(args.dir).resolve())

            if not (os.path.exists(abs_path) and os.path.isdir(abs_path)):
                errMsg = f"Specified path must exist and be a directory: {abs_path}"
                log.error(errMsg)
                return
            
            state.strategy_paths.append(abs_path)

            if args.clean:
                log.info(f"Starting clean, so will attempt to remove any existing caches!")
                cache_path = os.path.join(abs_path, '.cl_cache')
                if os.path.exists(cache_path):
                    try:
                        os.remove(cache_path)
                        log.info(f"Removed: {cache_path}")
                    except Exception as e:
                        log.error(f"Failed to remove {cache_path}!")
                        return
                    
    server = CLServer(state)
    register_endpoints(server)

    uvicorn.run(
        server,
        host=state.config.host,
        port=state.config.port,
        #reload=state.config.debug,
        log_level="info"
    )

def set_oneshot_arguments(parser):
    parser.add_argument("-c", "--clean", action="store_true", help="Start clean by disregarding any existing cache files.")
    parser.add_argument("-d", "--dir", type=str, required=True, help="Target directory. This takes precedence over `strat_config`.")
    parser.add_argument("--config", type=str, default="config/server_config.yaml", help="Server configuration YAML file.")
    parser.set_defaults(func=run_oneshot)

def run_oneshot(args):
    """ Run in oneshot mode """

    do_intro()

    try:
        config = ServerConfig.fromYAML(args.config)
    except Exception as e:
        print(f"Failed to load in server config: {str(e)}. Using defaults.")
        config = ServerConfig()
    
    do_oneshot(
        clean = args.clean,
        abs_path = args.dir,
        strat_config = config.strat_config("pyiml")
    )

if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="CodeLogician server", description="Run the CodeLogician server with various arguments.")
    parsers = parser.add_subparsers(title="subcommands", help="")

    # Oneshot parser
    oneshot_parser = parsers.add_parser ("oneshot", help="We should do a single run.")
    set_oneshot_arguments(oneshot_parser)

    # Server parser
    server_parser = parsers.add_parser("server", help="Start the server")
    set_server_arguments(server_parser)

    args = parser.parse_args()
    args.func(args)



"""Statistical operations package."""

from .basic import count_values, mean, sum_list
from .measures import standard_deviation, variance

__all__ = ["count_values", "mean", "standard_deviation", "sum_list", "variance"]



from pathlib import Path

from inline_snapshot import snapshot

from codelogician.dep_graph.file_tree import FileSystem
from codelogician.dep_graph.symbol_dep.python import build_py_dep_graph
from test_utils import find_pyproject


class TestResolveImport:
    def resolve_import(self, repo_name: str):
        data_dir = find_pyproject(Path(__file__)) / "data"
        py_repo_path = data_dir / repo_name
        fs = FileSystem.from_disk(py_repo_path, ["py"])
        symbol_dep = build_py_dep_graph(fs)
        # return "\n".join(sorted(symbol_dep.edges_repr().split("\n")))
        return symbol_dep.graph_repr()

    def test_sample_simple_math_lib(self):
        assert self.resolve_import("sample_simple_math_lib") == snapshot("""\
__init__.py
advanced/__init__.py
advanced/geometry/__init__.py
advanced/geometry/shapes.py
├── basic.py
└── utils/helpers.py
basic.py
math_ops.py
└── basic.py
utils/__init__.py
└── utils/helpers.py
utils/helpers.py
├── basic.py
└── math_ops.py
""")

    def test_sample_bank_app(self):
        assert self.resolve_import("sample_bank_app") == snapshot("""\
account.py
├── currency.py
└── transaction.py
asset.py
└── currency.py
bank.py
├── account.py
├── asset.py
├── currency.py
├── ledger.py
├── liability.py
└── transaction.py
currency.py
ledger.py
├── account.py
├── asset.py
├── currency.py
└── liability.py
liability.py
└── currency.py
transaction.py
└── currency.py
""")

    def test_sample_math_lib(self):
        assert self.resolve_import("sample_math_lib") == snapshot("""\
__init__.py
analysis/__init__.py
analysis/numerical/__init__.py
├── analysis/numerical/derivatives.py
└── analysis/numerical/integration.py
analysis/numerical/derivatives.py
└── core/arithmetic.py
analysis/numerical/integration.py
└── core/arithmetic.py
analysis/optimization/__init__.py
└── analysis/optimization/root_finding.py
analysis/optimization/root_finding.py
├── analysis/numerical/derivatives.py
├── core/arithmetic.py
└── core/comparison.py
core/__init__.py
├── core/arithmetic.py
└── core/comparison.py
core/advanced.py
├── core/arithmetic.py
└── core/comparison.py
core/arithmetic.py
core/comparison.py
stats/__init__.py
├── stats/basic.py
└── stats/measures.py
stats/basic.py
├── core/arithmetic.py
└── core/comparison.py
stats/measures.py
├── core/advanced.py
├── core/arithmetic.py
└── stats/basic.py
utils.py
├── core/advanced.py
├── core/arithmetic.py
├── stats/basic.py
└── stats/measures.py
""")



from datetime import datetime, timedelta
from .models import User, Project, Priority
from .storage import InMemoryStorage
from .services import TaskService


def run_cli():
    storage = InMemoryStorage()
    service = TaskService(storage)

    # Setup demo data
    user = User(id=1, name="Alice")
    project = Project(id=1, name="Sample Project")
    storage.add_user(user)
    storage.add_project(project)

    # Add a few tasks
    service.create_task(
        1, "Finish report", "Complete the quarterly report",
        Priority.HIGH, datetime.now() + timedelta(days=1),
        project_id=1, user_id=1
    )
    service.create_task(
        2, "Email client", "Send update to client",
        Priority.MEDIUM, datetime.now() - timedelta(days=2),
        project_id=1, user_id=1
    )

    print("== All Tasks ==")
    for t in storage.tasks.values():
        print(f"- {t.title} (priority: {t.priority.name}, deadline: {t.deadline}, completed: {t.completed})")

    print("\n== Overdue Tasks ==")
    for t in service.overdue_tasks():
        print(f"- {t.title} (deadline: {t.deadline})")

    print("\n== High Priority Tasks ==")
    for t in service.high_priority_tasks():
        print(f"- {t.title}")



#
#   Imandra Inc.
#
#   test_search.py
#

import unittest

from codelogician.strategy.model import Model
from codelogician.strategy.metamodel import MetaModel
from codelogician.strategy.cl_agent_state import Embedding

from imandra.u.agents.code_logician.command import (
    EditStateElementCommand,
    EmbedCommand,
    InitStateCommand,
)

from imandra.u.agents.code_logician.graph import GraphState

import random

class TestSearch(unittest.TestCase):
    """
    """

    def test_search_simple(self):
        """
        Test that MetaModel can correctly find models with precalculated embedding vectors
        """

        def rand_vector():
            return [ random.random() for i in range(3054)]

        m1e = Embedding(source="hello", vector = rand_vector() )
        m1 = Model(rel_path='one.py', src_code_embeddings=[m1e])

        m2e = Embedding(source="hello", vector = rand_vector() )
        m2 = Model(rel_path='two.py', src_code_embeddings=[m2e])

        m3e = Embedding(source="hello", vector = rand_vector() )
        m3 = Model(rel_path='three.py', src_code_embeddings=[m3e])

        models = { 'one.py': m1, 'two.py': m2, 'three.py': m3 }

        mmodel = MetaModel(src_dir_abs_path="data/code5", models=models)

        results = mmodel.search(rand_vector())

        for result in results:
            print (result)


    def test_search_call_cl_src_only(self):
        """
        Call CL to generate embeddings then insert them and do the search
        """
        pass

if __name__ == "__main__":
    unittest.main()


"""
Copy from networkx.readwrite.text.py and modified to make node repr customizable.

Text-based visual representations of graphs
"""

from __future__ import annotations

import sys
from collections import defaultdict
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import networkx as nx
from networkx.utils import open_file

if TYPE_CHECKING:
    from .base import DepGraph

__all__ = ["generate_network_text", "write_network_text"]


class BaseGlyphs:
    @classmethod
    def as_dict(cls):
        return {
            a: getattr(cls, a)
            for a in dir(cls)
            if not a.startswith("_") and a != "as_dict"
        }


class AsciiBaseGlyphs(BaseGlyphs):
    empty: str = "+"
    newtree_last: str = "+-- "
    newtree_mid: str = "+-- "
    endof_forest: str = "    "
    within_forest: str = ":   "
    within_tree: str = "|   "


class AsciiDirectedGlyphs(AsciiBaseGlyphs):
    last: str = "L-> "
    mid: str = "|-> "
    backedge: str = "<-"
    vertical_edge: str = "!"


class AsciiUndirectedGlyphs(AsciiBaseGlyphs):
    last: str = "L-- "
    mid: str = "|-- "
    backedge: str = "-"
    vertical_edge: str = "|"


class UtfBaseGlyphs(BaseGlyphs):
    # Notes on available box and arrow characters
    # https://en.wikipedia.org/wiki/Box-drawing_character
    # https://stackoverflow.com/questions/2701192/triangle-arrow
    empty: str = "╙"
    newtree_last: str = "╙── "
    newtree_mid: str = "╟── "
    endof_forest: str = "    "
    within_forest: str = "╎   "
    within_tree: str = "│   "


class UtfDirectedGlyphs(UtfBaseGlyphs):
    last: str = "└─╼ "
    mid: str = "├─╼ "
    backedge: str = "╾"
    vertical_edge: str = "╽"


class UtfUndirectedGlyphs(UtfBaseGlyphs):
    last: str = "└── "
    mid: str = "├── "
    backedge: str = "─"
    vertical_edge: str = "│"


def generate_network_text(
    graph,
    node_repr_fn: Callable[[Any], str] | None = None,
    with_labels=True,
    sources=None,
    max_depth=None,
    ascii_only=False,
    vertical_chains=False,
):
    """Generate lines in the "network text" format

    This works via a depth-first traversal of the graph and writing a line for
    each unique node encountered. Non-tree edges are written to the right of
    each node, and connection to a non-tree edge is indicated with an ellipsis.
    This representation works best when the input graph is a forest, but any
    graph can be represented.

    This notation is original to networkx, although it is simple enough that it
    may be known in existing literature. See #5602 for details. The procedure
    is summarized as follows:

    1. Given a set of source nodes (which can be specified, or automatically
    discovered via finding the (strongly) connected components and choosing one
    node with minimum degree from each), we traverse the graph in depth first
    order.

    2. Each reachable node will be printed exactly once on it's own line.

    3. Edges are indicated in one of four ways:

        a. a parent "L-style" connection on the upper left. This corresponds to
        a traversal in the directed DFS tree.

        b. a backref "<-style" connection shown directly on the right. For
        directed graphs, these are drawn for any incoming edges to a node that
        is not a parent edge. For undirected graphs, these are drawn for only
        the non-parent edges that have already been represented (The edges that
        have not been represented will be handled in the recursive case).

        c. a child "L-style" connection on the lower right. Drawing of the
        children are handled recursively.

        d. if ``vertical_chains`` is true, and a parent node only has one child
        a "vertical-style" edge is drawn between them.

    4. The children of each node (wrt the directed DFS tree) are drawn
    underneath and to the right of it. In the case that a child node has already
    been drawn the connection is replaced with an ellipsis ("...") to indicate
    that there is one or more connections represented elsewhere.

    5. If a maximum depth is specified, an edge to nodes past this maximum
    depth will be represented by an ellipsis.

    6. If a node has a truthy "collapse" value, then we do not traverse past
    that node.

    Parameters
    ----------
    graph : nx.DiGraph | nx.Graph
        Graph to represent

    with_labels : bool | str
        If True will use the "label" attribute of a node to display if it
        exists otherwise it will use the node value itself. If given as a
        string, then that attribute name will be used instead of "label".
        Defaults to True.

    sources : List
        Specifies which nodes to start traversal from. Note: nodes that are not
        reachable from one of these sources may not be shown. If unspecified,
        the minimal set of nodes needed to reach all others will be used.

    max_depth : int | None
        The maximum depth to traverse before stopping. Defaults to None.

    ascii_only : Boolean
        If True only ASCII characters are used to construct the visualization

    vertical_chains : Boolean
        If True, chains of nodes will be drawn vertically when possible.

    Yields
    ------
    str : a line of generated text

    Examples
    --------
    >>> graph = nx.path_graph(10)
    >>> graph.add_node("A")
    >>> graph.add_node("B")
    >>> graph.add_node("C")
    >>> graph.add_node("D")
    >>> graph.add_edge(9, "A")
    >>> graph.add_edge(9, "B")
    >>> graph.add_edge(9, "C")
    >>> graph.add_edge("C", "D")
    >>> graph.add_edge("C", "E")
    >>> graph.add_edge("C", "F")
    >>> nx.write_network_text(graph)
    ╙── 0
        └── 1
            └── 2
                └── 3
                    └── 4
                        └── 5
                            └── 6
                                └── 7
                                    └── 8
                                        └── 9
                                            ├── A
                                            ├── B
                                            └── C
                                                ├── D
                                                ├── E
                                                └── F
    >>> nx.write_network_text(graph, vertical_chains=True)
    ╙── 0
        │
        1
        │
        2
        │
        3
        │
        4
        │
        5
        │
        6
        │
        7
        │
        8
        │
        9
        ├── A
        ├── B
        └── C
            ├── D
            ├── E
            └── F
    """
    from typing import NamedTuple

    class StackFrame(NamedTuple):
        parent: Any
        node: Any
        indents: list
        this_islast: bool
        this_vertical: bool

    collapse_attr = "collapse"

    is_directed = graph.is_directed()

    if is_directed:
        glyphs = AsciiDirectedGlyphs if ascii_only else UtfDirectedGlyphs
        succ = graph.succ
        pred = graph.pred
    else:
        glyphs = AsciiUndirectedGlyphs if ascii_only else UtfUndirectedGlyphs
        succ = graph.adj
        pred = graph.adj

    if isinstance(with_labels, str):
        label_attr = with_labels
    elif with_labels:
        label_attr = "label"
    else:
        label_attr = None

    # <patch>
    def get_label(node) -> str:
        if node_repr_fn is not None:
            return node_repr_fn(node)
        if label_attr is not None:
            return str(graph.nodes[node].get(label_attr, node))
        else:
            return str(node)

    # </patch>

    if max_depth == 0:
        yield glyphs.empty + " ..."
    elif len(graph.nodes) == 0:
        yield glyphs.empty
    else:
        # If the nodes to traverse are unspecified, find the minimal set of
        # nodes that will reach the entire graph
        if sources is None:
            sources = _find_sources(graph)

        # Populate the stack with each:
        # 1. parent node in the DFS tree (or None for root nodes),
        # 2. the current node in the DFS tree
        # 2. a list of indentations indicating depth
        # 3. a flag indicating if the node is the final one to be written.
        # Reverse the stack so sources are popped in the correct order.
        last_idx = len(sources) - 1
        stack = [
            StackFrame(None, node, [], (idx == last_idx), False)
            for idx, node in enumerate(sources)
        ][::-1]

        num_skipped_children = defaultdict(lambda: 0)
        seen_nodes = set()
        while stack:
            parent, node, indents, this_islast, this_vertical = stack.pop()

            if node is not Ellipsis:
                skip = node in seen_nodes
                if skip:
                    # Mark that we skipped a parent's child
                    num_skipped_children[parent] += 1

                if this_islast:  # noqa: SIM102
                    # If we reached the last child of a parent, and we skipped
                    # any of that parents children, then we should emit an
                    # ellipsis at the end after this.
                    if num_skipped_children[parent] and parent is not None:
                        # Append the ellipsis to be emitted last
                        next_islast = True
                        try_frame = StackFrame(
                            node, Ellipsis, indents, next_islast, False
                        )
                        stack.append(try_frame)

                        # Redo this frame, but not as a last object
                        next_islast = False
                        try_frame = StackFrame(
                            parent, node, indents, next_islast, this_vertical
                        )
                        stack.append(try_frame)
                        continue

                if skip:
                    continue
                seen_nodes.add(node)

            if not indents:
                # Top level items (i.e. trees in the forest) get different
                # glyphs to indicate they are not actually connected
                if this_islast:
                    this_vertical = False
                    this_prefix = [*indents, glyphs.newtree_last]
                    next_prefix = [*indents, glyphs.endof_forest]
                else:
                    this_prefix = [*indents, glyphs.newtree_mid]
                    next_prefix = [*indents, glyphs.within_forest]

            else:
                # Non-top-level items
                if this_vertical:
                    this_prefix = indents
                    next_prefix = indents
                else:
                    if this_islast:
                        this_prefix = [*indents, glyphs.last]
                        next_prefix = [*indents, glyphs.endof_forest]
                    else:
                        this_prefix = [*indents, glyphs.mid]
                        next_prefix = [*indents, glyphs.within_tree]

            if node is Ellipsis:
                label = " ..."
                suffix = ""
                children = []
            else:
                # <patch>
                label = get_label(node)
                # if label_attr is not None:
                #     label = str(graph.nodes[node].get(label_attr, node))
                # else:
                #     label = str(node)
                # </patch>

                # Determine if we want to show the children of this node.
                if collapse_attr is not None:
                    collapse = graph.nodes[node].get(collapse_attr, False)
                else:
                    collapse = False

                # Determine:
                # (1) children to traverse into after showing this node.
                # (2) parents to immediately show to the right of this node.
                if is_directed:
                    # In the directed case we must show every successor node
                    # note: it may be skipped later, but we don't have that
                    # information here.
                    children = sorted(succ[node], key=str)  # Sort for determinism
                    # In the directed case we must show every predecessor
                    # except for parent we directly traversed from.
                    handled_parents = {parent}
                else:
                    # Showing only the unseen children results in a more
                    # concise representation for the undirected case.
                    children = sorted(
                        [child for child in succ[node] if child not in seen_nodes],
                        key=str,
                    )  # Sort for determinism

                    # In the undirected case, parents are also children, so we
                    # only need to immediately show the ones we can no longer
                    # traverse
                    handled_parents = {*children, parent}

                if max_depth is not None and len(indents) == max_depth - 1:
                    # Use ellipsis to indicate we have reached maximum depth
                    if children:
                        children = [Ellipsis]
                    handled_parents = {parent}

                if collapse:
                    # Collapsing a node is the same as reaching maximum depth
                    if children:
                        children = [Ellipsis]
                    handled_parents = {parent}

                # The other parents are other predecessors of this node that
                # are not handled elsewhere.
                other_parents = sorted(
                    [p for p in pred[node] if p not in handled_parents], key=str
                )  # Sort for determinism
                if other_parents:
                    # <patch>
                    other_parents_labels = [get_label(p) for p in other_parents]
                    other_parents_labels = ", ".join(other_parents_labels)
                    # if label_attr is not None:
                    #     other_parents_labels = ", ".join(
                    #         [
                    #             str(graph.nodes[p].get(label_attr, p))
                    #             for p in other_parents
                    #         ]
                    #     )
                    # else:
                    #     other_parents_labels = ", ".join(
                    #         [str(p) for p in other_parents]
                    #     )
                    # </patch>
                    suffix = " ".join(["", glyphs.backedge, other_parents_labels])
                else:
                    suffix = ""

            # Emit the line for this node, this will be called for each node
            # exactly once.
            if this_vertical:
                yield "".join([*this_prefix, glyphs.vertical_edge])

            yield "".join([*this_prefix, label, suffix])

            if vertical_chains:
                if is_directed:
                    num_children = len(set(children))
                else:
                    num_children = len(set(children) - {parent})
                # The next node can be drawn vertically if it is the only
                # remaining child of this node.
                next_is_vertical = num_children == 1
            else:
                next_is_vertical = False

            # Push children on the stack in reverse order so they are popped in
            # the original order.
            for idx, child in enumerate(children[::-1]):
                next_islast = idx == 0
                try_frame = StackFrame(
                    node, child, next_prefix, next_islast, next_is_vertical
                )
                stack.append(try_frame)


@open_file(1, "w")
def write_network_text(
    graph,
    path=None,
    with_labels=True,
    sources=None,
    max_depth=None,
    ascii_only=False,
    end="\n",
    vertical_chains=False,
):
    """Creates a nice text representation of a graph

    This works via a depth-first traversal of the graph and writing a line for
    each unique node encountered. Non-tree edges are written to the right of
    each node, and connection to a non-tree edge is indicated with an ellipsis.
    This representation works best when the input graph is a forest, but any
    graph can be represented.

    Parameters
    ----------
    graph : nx.DiGraph | nx.Graph
        Graph to represent

    path : string or file or callable or None
       Filename or file handle for data output.
       if a function, then it will be called for each generated line.
       if None, this will default to "sys.stdout.write"

    with_labels : bool | str
        If True will use the "label" attribute of a node to display if it
        exists otherwise it will use the node value itself. If given as a
        string, then that attribute name will be used instead of "label".
        Defaults to True.

    sources : List
        Specifies which nodes to start traversal from. Note: nodes that are not
        reachable from one of these sources may not be shown. If unspecified,
        the minimal set of nodes needed to reach all others will be used.

    max_depth : int | None
        The maximum depth to traverse before stopping. Defaults to None.

    ascii_only : Boolean
        If True only ASCII characters are used to construct the visualization

    end : string
        The line ending character

    vertical_chains : Boolean
        If True, chains of nodes will be drawn vertically when possible.

    Examples
    --------
    >>> graph = nx.balanced_tree(r=2, h=2, create_using=nx.DiGraph)
    >>> nx.write_network_text(graph)
    ╙── 0
        ├─╼ 1
        │   ├─╼ 3
        │   └─╼ 4
        └─╼ 2
            ├─╼ 5
            └─╼ 6

    >>> # A near tree with one non-tree edge
    >>> graph.add_edge(5, 1)
    >>> nx.write_network_text(graph)
    ╙── 0
        ├─╼ 1 ╾ 5
        │   ├─╼ 3
        │   └─╼ 4
        └─╼ 2
            ├─╼ 5
            │   └─╼  ...
            └─╼ 6

    >>> graph = nx.cycle_graph(5)
    >>> nx.write_network_text(graph)
    ╙── 0
        ├── 1
        │   └── 2
        │       └── 3
        │           └── 4 ─ 0
        └──  ...

    >>> graph = nx.cycle_graph(5, nx.DiGraph)
    >>> nx.write_network_text(graph, vertical_chains=True)
    ╙── 0 ╾ 4
        ╽
        1
        ╽
        2
        ╽
        3
        ╽
        4
        └─╼  ...

    >>> nx.write_network_text(graph, vertical_chains=True, ascii_only=True)
    +-- 0 <- 4
        !
        1
        !
        2
        !
        3
        !
        4
        L->  ...

    >>> graph = nx.generators.barbell_graph(4, 2)
    >>> nx.write_network_text(graph, vertical_chains=False)
    ╙── 4
        ├── 5
        │   └── 6
        │       ├── 7
        │       │   ├── 8 ─ 6
        │       │   │   └── 9 ─ 6, 7
        │       │   └──  ...
        │       └──  ...
        └── 3
            ├── 0
            │   ├── 1 ─ 3
            │   │   └── 2 ─ 0, 3
            │   └──  ...
            └──  ...
    >>> nx.write_network_text(graph, vertical_chains=True)
    ╙── 4
        ├── 5
        │   │
        │   6
        │   ├── 7
        │   │   ├── 8 ─ 6
        │   │   │   │
        │   │   │   9 ─ 6, 7
        │   │   └──  ...
        │   └──  ...
        └── 3
            ├── 0
            │   ├── 1 ─ 3
            │   │   │
            │   │   2 ─ 0, 3
            │   └──  ...
            └──  ...

    >>> graph = nx.complete_graph(5, create_using=nx.Graph)
    >>> nx.write_network_text(graph)
    ╙── 0
        ├── 1
        │   ├── 2 ─ 0
        │   │   ├── 3 ─ 0, 1
        │   │   │   └── 4 ─ 0, 1, 2
        │   │   └──  ...
        │   └──  ...
        └──  ...

    >>> graph = nx.complete_graph(3, create_using=nx.DiGraph)
    >>> nx.write_network_text(graph)
    ╙── 0 ╾ 1, 2
        ├─╼ 1 ╾ 2
        │   ├─╼ 2 ╾ 0
        │   │   └─╼  ...
        │   └─╼  ...
        └─╼  ...
    """
    if path is None:
        # The path is unspecified, write to stdout
        _write = sys.stdout.write
    elif hasattr(path, "write"):
        # The path is already an open file
        _write = path.write
    elif callable(path):
        # The path is a custom callable
        _write = path
    else:
        raise TypeError(type(path))

    for line in generate_network_text(
        graph,
        with_labels=with_labels,
        sources=sources,
        max_depth=max_depth,
        ascii_only=ascii_only,
        vertical_chains=vertical_chains,
    ):
        _write(line + end)


def _find_sources(graph):
    """
    Determine a minimal set of nodes such that the entire graph is reachable
    """
    # For each connected part of the graph, choose at least
    # one node as a starting point, preferably without a parent
    if graph.is_directed():
        # Choose one node from each SCC with minimum in_degree
        sccs = list(nx.strongly_connected_components(graph))
        # condensing the SCCs forms a dag, the nodes in this graph with
        # 0 in-degree correspond to the SCCs from which the minimum set
        # of nodes from which all other nodes can be reached.
        scc_graph = nx.condensation(graph, sccs)
        supernode_to_nodes = {sn: [] for sn in scc_graph.nodes()}
        # Note: the order of mapping differs between pypy and cpython
        # so we have to loop over graph nodes for consistency
        mapping = scc_graph.graph["mapping"]
        for n in graph.nodes:
            sn = mapping[n]
            supernode_to_nodes[sn].append(n)
        sources = []
        for sn in scc_graph.nodes():
            if scc_graph.in_degree[sn] == 0:
                scc = supernode_to_nodes[sn]
                node = min(scc, key=lambda n: graph.in_degree[n])
                sources.append(node)
    else:
        # For undirected graph, the entire graph will be reachable as
        # long as we consider one node from every connected component
        sources = [
            min(cc, key=lambda n: graph.degree[n])
            for cc in nx.connected_components(graph)
        ]
        sources = sorted(sources, key=lambda n: graph.degree[n])
    return sources


def _parse_network_text(lines):
    """Reconstructs a graph from a network text representation.

    This is mainly used for testing.  Network text is for display, not
    serialization, as such this cannot parse all network text representations
    because node labels can be ambiguous with the glyphs and indentation used
    to represent edge structure. Additionally, there is no way to determine if
    disconnected graphs were originally directed or undirected.

    Parameters
    ----------
    lines : list or iterator of strings
        Input data in network text format

    Returns
    -------
    G: NetworkX graph
        The graph corresponding to the lines in network text format.
    """
    from itertools import chain
    from typing import NamedTuple

    class ParseStackFrame(NamedTuple):
        node: Any
        indent: int
        has_vertical_child: int | None

    initial_line_iter = iter(lines)

    is_ascii = None
    is_directed = None

    ##############
    # Initial Pass
    ##############

    # Do an initial pass over the lines to determine what type of graph it is.
    # Remember what these lines were, so we can reiterate over them in the
    # parsing pass.
    initial_lines = []
    try:
        first_line = next(initial_line_iter)
    except StopIteration:
        ...
    else:
        initial_lines.append(first_line)
        # The first character indicates if it is an ASCII or UTF graph
        first_char = first_line[0]
        if first_char in {
            UtfBaseGlyphs.empty,
            UtfBaseGlyphs.newtree_mid[0],
            UtfBaseGlyphs.newtree_last[0],
        }:
            is_ascii = False
        elif first_char in {
            AsciiBaseGlyphs.empty,
            AsciiBaseGlyphs.newtree_mid[0],
            AsciiBaseGlyphs.newtree_last[0],
        }:
            is_ascii = True
        else:
            raise AssertionError(f"Unexpected first character: {first_char}")

    if is_ascii:
        directed_glyphs = AsciiDirectedGlyphs.as_dict()
        undirected_glyphs = AsciiUndirectedGlyphs.as_dict()
    else:
        directed_glyphs = UtfDirectedGlyphs.as_dict()
        undirected_glyphs = UtfUndirectedGlyphs.as_dict()

    # For both directed / undirected glyphs, determine which glyphs never
    # appear as substrings in the other undirected / directed glyphs.  Glyphs
    # with this property unambiguously indicates if a graph is directed /
    # undirected.
    directed_items = set(directed_glyphs.values())
    undirected_items = set(undirected_glyphs.values())
    unambiguous_directed_items = []
    for item in directed_items:
        other_items = undirected_items
        other_supersets = [other for other in other_items if item in other]
        if not other_supersets:
            unambiguous_directed_items.append(item)
    unambiguous_undirected_items = []
    for item in undirected_items:
        other_items = directed_items
        other_supersets = [other for other in other_items if item in other]
        if not other_supersets:
            unambiguous_undirected_items.append(item)

    for line in initial_line_iter:
        initial_lines.append(line)
        if any(item in line for item in unambiguous_undirected_items):
            is_directed = False
            break
        elif any(item in line for item in unambiguous_directed_items):
            is_directed = True
            break

    if is_directed is None:
        # Not enough information to determine, choose undirected by default
        is_directed = False

    glyphs = directed_glyphs if is_directed else undirected_glyphs

    # the backedge symbol by itself can be ambiguous, but with spaces around it
    # becomes unambiguous.
    backedge_symbol = " " + glyphs["backedge"] + " "

    # Reconstruct an iterator over all of the lines.
    parsing_line_iter = chain(initial_lines, initial_line_iter)

    ##############
    # Parsing Pass
    ##############

    edges = []
    nodes = []
    is_empty = None

    noparent = object()  # sentinel value

    # keep a stack of previous nodes that could be parents of subsequent nodes
    stack = [ParseStackFrame(noparent, -1, None)]

    for line in parsing_line_iter:
        if line == glyphs["empty"]:
            # If the line is the empty glyph, we are done.
            # There shouldn't be anything else after this.
            is_empty = True
            continue

        if backedge_symbol in line:
            # This line has one or more backedges, separate those out
            node_part, backedge_part = line.split(backedge_symbol)
            backedge_nodes = [u.strip() for u in backedge_part.split(", ")]
            # Now the node can be parsed
            node_part = node_part.rstrip()
            prefix, node = node_part.rsplit(" ", 1)
            node = node.strip()
            # Add the backedges to the edge list
            edges.extend([(u, node) for u in backedge_nodes])
        else:
            # No backedge, the tail of this line is the node
            prefix, node = line.rsplit(" ", 1)
            node = node.strip()

        prev = stack.pop()

        if node in glyphs["vertical_edge"]:
            # Previous node is still the previous node, but we know it will
            # have exactly one child, which will need to have its nesting level
            # adjusted.
            modified_prev = ParseStackFrame(
                prev.node,
                prev.indent,
                True,
            )
            stack.append(modified_prev)
            continue

        # The length of the string before the node characters give us a hint
        # about our nesting level. The only case where this doesn't work is
        # when there are vertical chains, which is handled explicitly.
        indent = len(prefix)
        curr = ParseStackFrame(node, indent, None)

        if prev.has_vertical_child:
            # In this case we know prev must be the parent of our current line,
            # so we don't have to search the stack. (which is good because the
            # indentation check wouldn't work in this case).
            ...
        else:
            # If the previous node nesting-level is greater than the current
            # nodes nesting-level than the previous node was the end of a path,
            # and is not our parent. We can safely pop nodes off the stack
            # until we find one with a comparable nesting-level, which is our
            # parent.
            while curr.indent <= prev.indent:
                prev = stack.pop()

        if node == "...":
            # The current previous node is no longer a valid parent,
            # keep it popped from the stack.
            stack.append(prev)
        else:
            # The previous and current nodes may still be parents, so add them
            # back onto the stack.
            stack.append(prev)
            stack.append(curr)

            # Add the node and the edge to its parent to the node / edge lists.
            nodes.append(curr.node)
            if prev.node is not noparent:
                edges.append((prev.node, curr.node))

    if is_empty:
        # Sanity check
        assert len(nodes) == 0

    # Reconstruct the graph
    cls = nx.DiGraph if is_directed else nx.Graph
    new = cls()
    new.add_nodes_from(nodes)
    new.add_edges_from(edges)
    return new


def test_original_gen_network_text():
    graph = nx.path_graph(10)
    graph.add_node("A")
    graph.add_node("B")
    graph.add_node("C")
    graph.add_node("D")
    graph.add_edge(9, "A")
    graph.add_edge(9, "B")
    graph.add_edge(9, "C")
    graph.add_edge("C", "D")
    graph.add_edge("C", "E")
    graph.add_edge("C", "F")
    assert (
        "\n".join(generate_network_text(graph))
        == """\
╙── 0
    └── 1
        └── 2
            └── 3
                └── 4
                    └── 5
                        └── 6
                            └── 7
                                └── 8
                                    └── 9
                                        ├── A
                                        ├── B
                                        └── C
                                            ├── D
                                            ├── E
                                            └── F\
"""
    )
    assert (
        "\n".join(generate_network_text(graph, vertical_chains=True))
        == """\
╙── 0
    │
    1
    │
    2
    │
    3
    │
    4
    │
    5
    │
    6
    │
    7
    │
    8
    │
    9
    ├── A
    ├── B
    └── C
        ├── D
        ├── E
        └── F\
"""
    )


def get_network_text(
    graph: DepGraph,
    node_repr_fn: Callable[[Any], str] | None = None,
    ascii_only: bool = True,
    **kwargs,
):
    if node_repr_fn is None:
        node_repr_fn = repr

    # Make output deterministic by sorting sources
    if "sources" not in kwargs:
        sources = _find_sources(graph.nx_graph)
        # Sort sources deterministically by string representation
        sources = sorted(sources, key=str)
        kwargs["sources"] = sources

    return "\n".join(
        generate_network_text(
            graph.nx_graph,
            node_repr_fn=node_repr_fn,
            ascii_only=ascii_only,
            **kwargs,
        )
    )



#
#   Imandra Inc.
#
#   cl_caller.py
#

from imandra.u.agents.code_logician.graph import GraphState
from imandra.u.agents.code_logician.command import InitStateCommand, GenFormalizationDataCommand, GenModelCommand
from imandra.u.agents.code_logician.base import FormalizationStatus
from imandra.u.agents import create_thread_sync, get_remote_graph
from queue import Queue
from threading import Thread
import asyncio

from ..strategy.cl_agent_state import CLAgentState

def clAutoformalize (src_code):
    """ Return the object containing results of running CodeLogician agent """

    graph = get_remote_graph("code_logician", api_key="29U5z4uV1E1Jbg6bOzdT4kpJUoqSKgaoaVzlGyt1zQfNXjFd")
    create_thread_sync(graph)

    # Create an empty agent state (`GraphState`), set the Python source program to the `src_code` element of the state and run the agent.
    gs = GraphState()

    # Create a list of commands for CL to execute
    gs = gs.add_commands([
        InitStateCommand (src_code=src_code, src_lang="python"), # Will initialize the state with specified Python source program
        GenFormalizationDataCommand(), # Will gather relevant formalization data (required to create the model)
        GenModelCommand(), # Will attempt to generate the formalized model
    ])

    res = asyncio.run(gs.run(graph)) # Run the agent
  
    f = res[0].last_fstate

    def formalizationStatusConverter(s):
        if s == FormalizationStatus.UNKNOWN:
            return "unknown"
        elif s == FormalizationStatus.INADMISSIBLE:
            return "inadmissable"
        elif s == FormalizationStatus.ADMITTED_WITH_OPAQUENESS:
            return "admitted_with_opaqueness"
        elif s == FormalizationStatus.EXECUTABLE_WITH_APPROXIMATION:
            return "executable_with_approximation"
        elif s == FormalizationStatus.TRANSPARENT:
            return "transparent"
        else:
            raise Exception (f"Something went wrong!!!")

    return CLAgentState (
        status = formalizationStatusConverter(f.status),
        src_code = f.src_code,
        iml_code = f.iml_code,
        vgs = f.vgs,
        region_decomps = f.region_decomps,
        opaque_funcs=f.opaques
    )



from basic import add, multiply
from math_ops import square


def sum_three(a: float, b: float, c: float) -> float:
    """Add three numbers."""
    return add(add(a, b), c)


def max_of_three(a: float, b: float, c: float) -> float:
    """Find maximum of three numbers."""
    if a >= b and a >= c:
        return a
    elif b >= c:
        return b
    else:
        return c


def circle_area(radius: float) -> float:
    """Calculate area of circle."""
    pi = 3.14159
    return multiply(pi, square(radius))


def is_even(n: int) -> bool:
    """Check if integer is even."""
    return n % 2 == 0



#
#   Imandra Inc.
#
#   javaiml_strategy.py
#

from .pyiml_strategy import PyIMLStrategy
from .state import StrategyState
from ..server.events import (
    ServerEvent,
    FileSystemCheckbackEvent,
    FileSystemEvent
)
import logging

log = logging.getLogger(__name__)

class JavaIMLStrategy(PyIMLStrategy):
    """ Java strategy  """

    language = 'Java'

    def __init__ (self, state:StrategyState):
        super().__init__(state=state)

    def watch_directories(self):
        """ Return the list of directories the observer should watch """
        return []
        #return self.abs_src_dir

    def on_load (self):
        """ What should be done on startup """
        pass

    def on_filesystem_event(self, event : FileSystemEvent):
        """ """
        log.info(f"Received {event} to process")
    
    def on_save(self):
        """
        What should be done on strategy save
        """
        pass





#
# Imandra Inc.
#
# metamodel_endpoints.py
#


from fastapi import HTTPException
from .cl_server import CLServer
import logging

log = logging.getLogger(__name__)

def register_metamodel_endpoints(app : CLServer):

    # Return summary of the current state
    @app.get("/metamodel/{strategy_id}/latest/summary", operation_id="get_summary")
    async def metamodel_latest_summary(strategy_id:str):
        """
        Return summary of the latest metamodel
        """

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(status_code=404, detail=f"Failed to get strategy state: {e}")

        log.info("Received /state/summary request")
        # already returns JSON representation
        if state.curr_meta_model is None:
            return {}
        else:
            return state.curr_meta_model.summary()

    @app.get("/metamodel/{strategy_id}/latest/list", operation_id="get_models_list")
    async def model_list(strategy_id:str, listby:str="frm_status"):
        """
        Return a list of model statistics, sorted by specified criteria. Options are:
        - frm_status - formalization status
        - upstream - number of models upstream which are affected by the specified model
        - opaques - number of opaque functions
        - failed_vgs - number of failed verification goals
        """
        log.info(f"Received request for model listing by {listby}.")

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(status_code=404, detail=f"Failed to get strategy state: {e}")

        try:
            if state.curr_meta_model is None:
                return []
            else:
                return state.curr_meta_model.gen_listing(listby)
        except Exception as e:
            return HTTPException(status_code=404, detail=f"Error when requesting listing: {str(e)}")

    @app.get("/metamodel/{strategy_id}/latest/vgs", operation_id="list_of_vgs")
    async def verification_goals(strategy_id:str):
        """
        Return verification goals with their statuses for the entire metamodel
        """

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(status_code=404, detail=f"Failed to get strategy state: {e}")
  
        log.info("Received /state/latest/vgs request")
        if state.curr_meta_model is None: return []
        
        return state.curr_meta_model.vgs()

    @app.get("/metamodel/{strategy_id}/latest/decomps", operation_id="list_of_decomps")
    async def decomps(strategy_id:str):
        """
        Return the list of decomps
        """

        log.info("Received /state/latest/decomps request")

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(status_code=404, detail=f"Failed to get strategy state: {e}")

        if state.curr_meta_model is None: return []
        return state.curr_meta_model.decomps()

    @app.get("/metamodel/{strategy_id}/latest/opaques", operation_id="list_of_opaque_functions")
    async def opaques(strategy_id:str):
        """
        Return list of opaque functions
        """

        log.info("Received /state/latest/opaques request")

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(status_code=404, detail=f"Failed to get strategy state: {e}")

        if state.curr_meta_model is None: return []
        return state.curr_meta_model.opaques()

    @app.get("/metamodel/{strategy_id}/latest", operation_id="latest_metamodel")
    async def latest_state (strategy_id:str):
        """
        Return the latest metaModel
        """
        log.info("Received request for the latest metamodel")

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(status_code=404, detail=f"Failed to get strategy state: {e}")

        if state.curr_meta_model is None: return {}
        return state.curr_meta_model.toJSON()

    @app.get("/metamodel/{strategy_id}/cache/by_index/{idx}", operation_id="get_metamodel_from_cache")
    async def get_state(strategy_id:str, idx : int):
        """
        Returns a complete state of the cache.
        """
        log.info("Received cache metamodel request for index {idx}")

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(status_code=404, detail=f"Failed to get strategy state: {e}")

        if idx in state.meta_cache.indices():
            mmodel = state.meta_cache.get_cache_metamodel(idx)
            if mmodel is None:
                return HTTPException(status_code=404, detail=f"Metamodel from cache with index={idx} not found!")
            else:
                return mmodel.toJSON()
        else:
            log.error(f"{idx} index not found!")
            return HTTPException(status_code=404, detail=f"Metamodel from cache with index={idx} not found!")

    @app.get("/metamodel/{strategy_id}/cache/indices", operation_id="get_metamodel_cache_indices")
    async def cache_indices (strategy_id:str) -> list[int]:
        """
        Return the list of all indices in cache.
        """

        log.info("Request for all indices.")

        try:
            state = app.strategy_state_by_id(strategy_id)
        except Exception as e:
            raise HTTPException(status_code=404, detail=f"Failed to get strategy state: {e}")
    
        return state.meta_cache.indices()



#
#   Imandra Bank
#
#   currency.py
#


class CurrencySymbol:
    EUR = 1
    USD = 2
    GBP = 3
    JPY = 4
    CNY = 5
    INR = 6
    AUD = 7
    CAD = 8
    CHF = 9
    NZD = 10


class CurrencyPairs:
    """CurrencyPairs class to represent a collection of currency pairs"""

    def __init__(self, pairs):
        """Initialize with a dictionary of currency pairs"""

        self._pairs = pairs

    def convert(self, from_currency: int, to_currency: int, amount: float) -> float:
        """Convert an amount from one currency to another"""

        if from_currency == to_currency:
            return amount

        pair_key = (from_currency.name, to_currency.name)
        if pair_key not in self._pairs:
            raise ValueError(f"Currency pair {pair_key} not found")

        rate = self._pairs[pair_key]
        return amount * rate

    def __repr__(self):
        """String representation of the currency pairs"""
        return ""


if __name__ == "__main__":
    rates = {
        (CurrencySymbol.EUR, CurrencySymbol.USD): 1.1,
        (CurrencySymbol.USD, CurrencySymbol.EUR): 0.9,
        (CurrencySymbol.GBP, CurrencySymbol.USD): 1.3,
        (CurrencySymbol.USD, CurrencySymbol.GBP): 0.77,
        (CurrencySymbol.JPY, CurrencySymbol.USD): 0.009,
        (CurrencySymbol.USD, CurrencySymbol.JPY): 110.0,
        (CurrencySymbol.CNY, CurrencySymbol.USD): 0.15,
        (CurrencySymbol.USD, CurrencySymbol.CNY): 6.5,
        (CurrencySymbol.INR, CurrencySymbol.USD): 0.013,
        (CurrencySymbol.USD, CurrencySymbol.INR): 75.0,
        (CurrencySymbol.AUD, CurrencySymbol.USD): 0.7,
        (CurrencySymbol.USD, CurrencySymbol.AUD): 1.4,
        (CurrencySymbol.CAD, CurrencySymbol.USD): 0.8,
        (CurrencySymbol.USD, CurrencySymbol.CAD): 1.25,
        (CurrencySymbol.CHF, CurrencySymbol.USD): 1.05,
        (CurrencySymbol.USD, CurrencySymbol.CHF): 0.95,
        (CurrencySymbol.NZD, CurrencySymbol.USD): 0.65,
        (CurrencySymbol.USD, CurrencySymbol.NZD): 1.54,
    }

    pairs = CurrencyPairs(rates)

    pairs.convert(CurrencySymbol.EUR, CurrencySymbol.USD, 100)  # Should return 110.0
    pairs.convert(CurrencySymbol.USD, CurrencySymbol.EUR, 100)  # Should return 90.0
    pairs.convert(CurrencySymbol.GBP, CurrencySymbol.USD, 100)  # Should return 130.0
    pairs.convert(CurrencySymbol.USD, CurrencySymbol.GBP, 100)  # Should return 77.0
    pairs.convert(CurrencySymbol.JPY, CurrencySymbol.USD, 100)  # Should return 0.9
    pairs.convert(CurrencySymbol.USD, CurrencySymbol.JPY, 100)  # Should return 11000.0



#
#   Imandra Inc.
#
#   model.py
#

from imandra.u.agents.code_logician.base import (
    FormalizationDependency,
    FormalizationStatus,
    ModuleInfo
)

from typing import Optional, Dict
from pathlib import Path
import datetime, logging

from rich.text import Text
from rich.pretty import Pretty
from rich.panel import Panel
from rich import print as rprint

from pydantic import BaseModel, field_serializer
import numpy as np

log = logging.getLogger(__name__)

from .model_status import ModelStatus, SrcCodeStatus
from .model_task import (
    ModelTask, UserManualIMLEditTask, FormalizationNeed
)
from .cl_agent_state import CLAgentState, Embedding
from ..tools.cl_caller import clAutoformalize


class Model(BaseModel):
    """
        Container for a model - representing a single model along with its dependendencies, 
        formalization state, etc...
        
        These are the primary actions for a model:
        - 1. Update the source code
        - 2. Reset the source code to the formalization state -

        Context - Human-provided information to CodeLogician for this model. If it's set, then
        `NeedsFormalization` function will return True:

        AgentState contains the latest results of running CodeLogician on this model
        - 5. Apply agentState - this sets the agent state (formalization state)

        Three ways that a status is accessed:
        - needsFormalization(timeToWaitSinceSrcCodeChange : int) -> bool -- returns True if the model
        requires updated formalization
        -- hasSrcCodeChanged(timeToWaitSinceSrcCodeChange : int) -> bool
        -- depsChanged() -> bool - models in the dependencies list have changed
        -- instructionsPresent() -> bool: human instructions have been provided

        This function is used during autoformalization to determine that the order
        of model formalization (if there're "blocking" dependencies, then we will not)
        - deps_need_formalization() -> bool - if this is True, then metamodel will not

        Then there are other functions like:
        - 1. Accessing/modifying lists of dependencies (other models that this model depends on)
        - 2. Saving to / loading from JSON files
        - 3. Generating ModelTask (object with instructions to CodeLogician)
    """

    rel_path : str
    src_code : Optional[str] = None
    src_code_last_changed : Optional[datetime.datetime] = None
    src_language : str = "Python"
    agent_state : Optional[CLAgentState] = None

    # Here we account for user's changes to the model IML code 
    # and keep track of how long it's been since they edited it
    # if it's been long enough and it's different that what we have
    # then we'll kick off a command to make it part of the Graph State
    user_iml_code : Optional[str] = None # This is human provided IML code
    user_iml_code_last_changed : Optional[datetime.datetime] = None

    src_code_embeddings : list[Embedding] = [] # Embeddings for the original source code
    iml_code_embeddings : list[Embedding] = [] # IML code embeddings

    # This is human-provided context that is sent to the CL agent
    context : Optional[str] = None

    # This is used to keep track of the latest TaskID that was generated
    # We can only apply changes to the CL agent state if the taskID of the
    # result andf what we're waiting for matches - otherwise we discard the results
    outstanding_task_ID : Optional[str] = None

    # This will be typically inserted after creation of the object because
    # this requires actual object references (to other models)
    dependencies : list = [] # list of models that this model depends on
    rev_dependencies : list = [] # list of models that this model affects (the inverse of dependencies)

    @field_serializer('dependencies','rev_dependencies')
    def serialize_tags(self, deps: list) -> list:
        return list (map (lambda x: x.rel_path, deps))

    # This gets updated along with each agent state - here we maintain
    # IML code for the dependencies that were used. If they differ
    # then we may need to re-run CodeLogician.
    formalized_deps : list[FormalizationDependency] = []

    # We do the same with the human-provided context
    formalized_context : Optional[str] = None

    def add_dependency(self, d) -> None:
        """
        "Safe" add of a dependency - if it's already there, it would not add it twice
        """
        if d in self.dependencies: return
        self.dependencies.append(d)

    def status(self) -> ModelStatus:
        """ 
        Return a status object that contains various bits of the current state of the model
        """

        return ModelStatus(
            src_code_status=self.src_code_status(),
            instructions_added=self.context_provided(),
            deps_changed=self.deps_changed(),
            deps_need_formalization=self.deps_need_formalization(),
            formalization_status=self.formalization_status()
        )

    @staticmethod
    def calc_distance (vec1, vec2):
        """
        Calculate cosine distance between two vectors
        """

        dot_product = np.dot(vec1, vec2)

        norm_vec1 = np.linalg.norm(vec1)
        norm_vec2 = np.linalg.norm(vec2)

        if norm_vec1 == 0 or norm_vec2 == 0:
            return 1.0  # Handle cases where one or both vectors are zero vectors
        
        cosine_similarity = dot_product / (norm_vec1 * norm_vec2)
        cosine_distance = 1 - cosine_similarity
        return cosine_distance

    def calc_embedding_distance (self, query_vector : list[float]) -> float | None:
        """
        Calculate distance between the specified query vector and its own embeddings. Return the most minimal one.
        """

        qv = np.array(query_vector)

        min_distance = None
        for embedding in (self.iml_code_embeddings + self.src_code_embeddings):            
            curr_vec = np.array(embedding.vector)

            try:
                distance = self.calc_distance(qv, curr_vec)
            except Exception as e: raise Exception(f"Failed to calculate distance: {e}")

            if min_distance is None:
                min_distance = distance
            else:
                if distance < min_distance:
                    min_distance = distance
        
        if min_distance is None:
            return None
        return float(min_distance)

    def needs_embedding_update(self) -> bool:
        """
        Does this model need to update embedding info?
        """
        # TODO check to see if the source code/etc changed since last embedding...
        return False

    def gen_embedding_task (self) -> ModelTask:
        """
        Create an embedding task - TODO: we should streamline this 
        """
        task = ModelTask(
            rel_path=self.rel_path,
            src_code=self.src_code if self.src_code else "",
            context=self.context,
            gen_embeddings=True
        )

        # Let's now save this task so we know that we should expect 
        # the result 
        self.outstanding_task_ID = task.task_id

        return task

    def gen_formalization_task(self) -> ModelTask:
        """
        Generate a formalization taks (ModelTask) object if we need to
        """

        task = ModelTask(
            rel_path=self.rel_path,
            src_code=self.src_code if self.src_code else "",
            context=self.context,
            graph_state=self.agent_state.graph_state if self.agent_state else None,
            dependencies=self.depdendencies_iml_models()
        )

        self.outstanding_task_ID = task.task_id

        return task
    
    def gen_iml_user_update_task (self) -> ModelTask:
        """
        Generate a model task for updating the IML code within the
        """

        if self.user_iml_code is None:
            return None

        task = UserManualIMLEditTask(
            rel_path=self.rel_path,
            graph_state = self.agent_state.graph_state if self.agent_state else None,
            iml_code=self.user_iml_code
        )

        self.outstanding_task_ID = task.task_id

        return task

    def context_provided(self) -> bool:
        """ 
        Return True if there's human-provided context 
        """
        return bool(self.context)
  
    def apply_agent_state (self,
        agentState : CLAgentState,
        dependencies : list[FormalizationDependency] = []) -> None:
        """
        Apply the agent state
        """

        # Let's make sure we have an actual CLAgentState
        if not isinstance(agentState, CLAgentState):
            raise Exception (f"Expected a value of CLAgentState, but got something else: {str(agentState)}!")

        # Agent state
        self.agent_state = agentState

        # Remove instructions that were used from the current list of outstanding ones
        self.formalized_context = agentState.context

        # Let's now set the dependencies' models that were used
        self.formalized_deps = dependencies

        self.src_code_embeddings = agentState.src_code_embeddings
        self.iml_code_embeddings = agentState.iml_code_embeddings

    def user_iml_change_ready(self, user_wait_time:int) -> bool:
        """
        Has enough time passed since user modified the IML model?
        """

        if self.user_iml_code is None:
            return False

        # if we're missing formalization altogether, then source code has changed
        if self.agent_state is None:
            return True

        if user_wait_time is None or self.src_code_last_changed is None:
            timeLongEnough = True
        else:
            timeLongEnough = (datetime.datetime.now() - self.src_code_last_changed) > datetime.timedelta(seconds=user_wait_time)

        return timeLongEnough and (hash(self.user_iml_code) != hash(self.agent_state.iml_code))

    def formalization_reasons(self, user_wait_time:Optional[int]=None) -> list[FormalizationNeed]:
        """
        Returns a list of reasons for the need to perform formalization on this model.
        If the list is empty - then there's no need to do it.
        """

        if self.src_code is None: return [] # if we have no source code (i.e. file was deleted), then there's nothing to do    
        if self.agent_state is None: return [FormalizationNeed.NO_AGENT_STATE]

        reasons = []
        if self.has_src_code_changed(user_wait_time): reasons.append(FormalizationNeed.SRC_CODE_CHANGED)
        if self.context_provided(): reasons.append(FormalizationNeed.CONTEXT_ADDED)
        if self.deps_changed(): reasons.append(FormalizationNeed.DEPS_CHANGED)

        return reasons

    def deps_need_formalization(self, srcWaitTime:int=0):
        """ 
        Return True if there're dependencies (may be indirect) that are due for formalization and
        False otherwise. We use this function when constructing the list of Tasks for the metamodel. 
        """

        for d in self.dependencies:
            if d.formalization_reasons(srcWaitTime) or d.deps_need_formalization(srcWaitTime):
                return True

        return False

    def depdendencies_iml_models(self) -> list[FormalizationDependency]:
        """
        Return IML code of the dependencies
        """

        ds = []
        for model in self.dependencies:
            # We only care about the models that have IML code available
            if model.iml_code() is None: continue

            src_mod_info = model.to_CL_src_module_info()
            iml_mod_info = model.to_CL_iml_module_info()

            ds.append(
                FormalizationDependency(
                src_module=src_mod_info,
                iml_module=iml_mod_info,
                )
            )

        return ds

    def to_CL_src_module_info(self) -> ModuleInfo:
        """
        Create ModuleInfo
        """

        return ModuleInfo(
            name = self.rel_path,
            relative_path=Path(self.rel_path),
            content = self.src_code if self.src_code else "",
            src_lang = self.src_language
        )

    def to_CL_iml_module_info(self) -> ModuleInfo | None:
        """
        Return ModuleInfo object for the IML code if it's available, None otherwise
        """

        # If we don't have any IML code for this model, let's not return it
        if self.iml_code() is None:
            return None
        
        if self.iml_code() is not None:
            iml_code = str(self.iml_code())
        else:
            iml_code = ""

        return ModuleInfo (
            name = self.rel_path,
            relative_path = Path(self.rel_path),
            content = iml_code,
            src_lang="IML"
        )

    def src_code_status(self, srcCodeWaitTime:Optional[int]=None) -> SrcCodeStatus:
        """
        Return source code status
        """

        if self.src_code is None:
            return SrcCodeStatus.SRC_CODE_DELETED

        else:
            if srcCodeWaitTime is None or self.src_code_last_changed is None:
                timeLongEnough = True
            else:
                timeLongEnough = (datetime.datetime.now() - self.src_code_last_changed) > datetime.timedelta(seconds=srcCodeWaitTime)

        if timeLongEnough and (hash(self.src_code) == hash(self.iml_code())):
            return SrcCodeStatus.SRC_CODE_CHANGED
        else:
            return SrcCodeStatus.SRC_CODE_CURRENT

    def formalization_status(self) -> FormalizationStatus:
        """
        Return formalization status of the model
        """
        if self.agent_state is None:
            return FormalizationStatus.UNKNOWN
        else:
            return self.agent_state.status

    def has_src_code_changed(self, srcCodeWaitTime:Optional[int]=None) -> bool:
        """
        Return True if source code has changed from when it was formalized
        """

        # if we're missing formalization altogether, then source code has changed
        if self.agent_state is None:
            return True

        if srcCodeWaitTime is None or self.src_code_last_changed is None:
            timeLongEnough = True
        else:
            timeLongEnough = (datetime.datetime.now() - self.src_code_last_changed) > datetime.timedelta(seconds=srcCodeWaitTime)

        return timeLongEnough and (hash(self.src_code) != hash(self.agent_state.src_code))

    def deps_changed(self) -> bool:
        """    
        Return True if dependencies' models (IML code) have changed since we
        ran CL formalization task last time. When we receive CL results, we always
        record dependencies' IML models that were used. 
        """

        currentDeps = self.depdendencies_iml_models()

        if len(self.formalized_deps) != len(currentDeps):
            return True

        for d in self.formalized_deps:
            if d not in currentDeps:
                return True

        return False

    def set_iml_model (self, new_iml_model : str, record_time:bool = False) -> None:
        """
        Set the IML model code and record the time if we need to. This is used
        for users when manually overriding the model.
        """

        if not isinstance(new_iml_model, str):
            raise Exception (f"Expected a string for 'new_iml_model', but got {type(new_iml_model).__name__ }")

        #if hash(new_iml_model) != hash(self.iml_code()):
        if hash(new_iml_model) != hash(self.user_iml_code):
            log.info(f"User is updating the IML model: [{self.rel_path}]")

            if record_time:
                self.user_iml_code_last_changed = datetime.datetime.now()
            else:
                self.user_iml_code_last_changed = None
            
            self.user_iml_code = new_iml_model
        else:
            log.warning(f"Model set by the user doesn't appear to be any different from existing model: [{self.rel_path}]")

    def set_src_code(self, new_src_code : str|None, record_time:bool = False) -> None:
        """
        Set the src_code and update the time source code changed
        """

        # This means the code was deleted
        if new_src_code is None:
            self.src_code = None
            self.src_code_last_changed = None
            return 

        if not isinstance(new_src_code, str):
            raise Exception (f"Expected a string value for `new_src_code`, but got {type(new_src_code).__name__}")

        if hash(new_src_code) != hash(self.src_code):
            log.info(f"Updating source code for model [{self.rel_path}]")
      
            if record_time:
                self.src_code_last_changed = datetime.datetime.now()
            else:
                self.src_code_last_changed = None

            self.src_code = new_src_code
        else:
            log.warning(f"Specified model hasn't changed: [{self.rel_path}]")


    def iml_code(self) -> str|None:
        """
        If available, returns IML code with artifacts (e.g. VGs and decomps)
        """
        return self.agent_state.iml_code if self.agent_state else None

    def iml_model(self) -> str|None:
        """
        If available, returns IML model (not artifacts like VGs, decomps) 
        """
        return self.agent_state.iml_model if self.agent_state else None

    def verification_goals(self) -> list:
        """
        If available, returns the list of dictionaries with verification goals
        """
        return self.agent_state.vgs if self.agent_state else []

    def failed_vgs(self) -> list:
        """
        Return just the list of failed VGs
        """
        return list(filter(lambda x: x['status'] == False, self.verification_goals()))

    def decomps(self) -> list:
        """
        If available, returns the list of decomposition requests
        """
        return self.agent_state.region_decomps if self.agent_state else []
    
    def opaque_funcs (self) -> list:
        """
        If available, return the opaque functions used along with their approximations
        """
        return self.agent_state.opaque_funcs if self.agent_state else []

    def gen_stats(self) -> Dict:
        """
        Generate some numberical stats for this model
        """

        s = {}
        s['frm_status'] = str(self.formalization_status())
        s['num_opaques'] = len(self.opaque_funcs())
        s['num_failed_vgs'] = len(self.failed_vgs())

        return s

    def toJSON(self) -> str:
        """
        Return a dictionary we can save to disk
        """
        return self.model_dump_json()

    @staticmethod
    def fromJSON(j : str|dict) -> 'Model':
        """
        Return a Model object from provided JSON. Note that we do not set the lists
        of models for 'dependsOn' here because we don't have access to the
        actual model objects. This is done outside this state methods.
        """

        if isinstance(j, str):
            return Model.model_validate_json(j)
        elif isinstance(j, dict):
            return Model.model_validate(j)
        else:
            raise Exception (f"Input must be either a str or a dict!")

    def str_summary(self) -> str:
        """
        Return a one-line string with high-level details
        """

        return f"[status={self.status()}]"

    def __hash__ (self):
        """
        We need this so we can compare models
        """
        return hash(str(self.toJSON()))

    def __repr__(self):
        """
        Return a nice representation
        """

        dependencyPaths = list(map(lambda x: x.rel_path, self.dependencies))
        agentStateStr = "None" if self.agent_state is None else self.agent_state.status

        s  = f"Model: {self.rel_path} \n"
        s += f"{str(self.status())}\n"
        s += f"Formalization state: {agentStateStr}\n"
        s += f"Depends on: {dependencyPaths}\n"
        s += f"Source language: {self.src_language}\n"

        if self.src_code:
            s += f"Source code (condensed): \n {self.src_code[:100]}\n"
        else:
            s += f"Source code (condensed): \n None \n"

        s += f"Opaque funcs: {len(self.opaque_funcs())}\n"
        s += f"Decomps: {len(self.decomps())}\n"
        return s

    def __rich__ (self):
        """ Return a Rich representation """

        pretty = Pretty(self.toJSON(), indent_size=2)
        return Panel(pretty, title="Model", border_style="green")



#
#   Imandra Inc.
#
#   test_pyiml_strategy.py
#

import unittest
from codelogician.strategy.state import StrategyState
from codelogician.strategy.pyiml_strategy import PyIMLStrategy, StratMode
from codelogician.strategy.events import ChangeMode


class TestPyIMLStrategy(unittest.TestCase):
    """ 
    Test PyIML strategy worker
    """

    def test_codelogician_calls(self):
        """ """

        print (f"Step 1: Load in the state value")
        state = StrategyState.from_directory("data/code2")

        print (state.curr_meta_model)

        for m in state.curr_meta_model.models:
            print (m)

        print(f"Step 2: Start the thread")
        strat = PyIMLStrategy(state)
        strat.start()

        print (f"Step 3: Send an event to the strategy - this should kick off autoformalization for us")
        # All files in the directory
        strat.add_event (ChangeMode(new_mode=StratMode.AUTO))

        print (f"Step 4: Wait until it completes")
        import time
        time.sleep(30)

        strat.add_event(None)

        print (f"Step 5: Print out the summary and the models")

        print (f"Step 6: Our models should now have mode set to `transparent`")
        #for path in models:
        #  #print (models[path])

        state.save()

if __name__ == "__main__":
  unittest.main()



#
#   Imandra Inc.
#
#   sketch.py
#    

import unittest, string, random, uuid
from pathlib import Path
from rich.text import Text
from rich.pretty import Pretty
from rich.panel import Panel
from rich import print as rprint

from enum import StrEnum
from pydantic import BaseModel, Field
from typing import Optional, Dict, Any

from .model import Model
from .sketch_task import (
    SketchChangeTask, SketchChangeResult,
    SketchChgSetModel, SketchChgInsertDef, 
    SketchChgModifyDef, SketchChgDeleteDef,
    SketchChange
)
from imandra.u.agents.code_logician.base.region_decomp import RegionDecomp
from imandra.u.agents.code_logician.base.vg import VG
from imandra.u.agents.code_logician.base import FormalizationStatus
#from imandra.u.agents.code_logician.base.imandrax import DecomposeRes, EvalRes, PO_Res
from ..tools.iml_utils import (
    add_definition, remove_definition, update_definition
)

class VGChange(BaseModel):
    """
    Helps us track VG changes over different states
    """
    is_new : bool # Is this a new VG that was just added?
    is_removed : bool # Has this VG been removed?
    res_changed : Optional[tuple[str,str]] = None # old vs new result type
    vg : VG # Actual VG info

class ChangeSummary(BaseModel):
    """
    Summary of the formalization state changes since the last state where we had an admissable model
    """

    vg_changes : list[VGChange]

    def stats(self) -> Dict[str, int]:
        """
        Return VG change statistics
        """

        def status_to_int (s:str) -> int:
            """
            """
            if s == 'unknown':
                return 1
            elif s == 'error' or s == 'err':
                return 2
            elif s == 'refuted':
                return 3
            elif s == 'verified_upto':
                return 4
            elif s == 'proved':
                return 5
            else:
                raise Exception (f"Unknown VG status: >{s}<")

        added, removed = 0, 0
        improved, deteriorated = 0, 0
        proved, refuted = 0, 0

        for vg_change in self.vg_changes:
            if vg_change.is_new:
                added += 1

            if vg_change.is_removed:
                removed += 1
                continue

            if vg_change.res_changed is None: continue

            if vg_change.res_changed[1] in ['proved', 'verified_upto']: proved += 1
            if vg_change.res_changed[1] in ['refuted']: refuted += 1

            if not vg_change.is_new:
                if status_to_int(vg_change.res_changed[0]) > status_to_int(vg_change.res_changed[1]):
                    deteriorated += 1
                elif status_to_int(vg_change.res_changed[0]) > status_to_int(vg_change.res_changed[1]):
                    improved += 1

        return {
            'added': added, 
            'removed': removed, 
            'improved': improved, 
            'deteriorated': deteriorated, 
            'proved': proved, 
            'refuted': refuted
        }

def calc_vg_difference(old_vgs : list[VG], new_vgs : list[VG]) -> list[VGChange]:
    """
    Compute difference between two lists of VGs
    """

    vg_diff : list[VGChange] = []

    def find_a_match(new_v:VG, other_vgs:list[VG]) -> VG | None:
        if new_v.data is None: return None
        return next(filter(lambda x: (x.data is not None) and (x.data.predicate == new_v.data.predicate), other_vgs), None) # pyright: ignore

    for new_vg in new_vgs:
        if new_vg.res is None: continue

        existing_vg = find_a_match(new_vg, old_vgs)

        if existing_vg is None:
            vg_diff.append (
                VGChange (
                    is_new=True, 
                    is_removed=False, 
                    res_changed=("", new_vg.res.res_type),
                    vg=new_vg
                )
            )
        else:
            # This is an existing VG, let's see if its result type has changed at all...
            if existing_vg.res is not None and new_vg.res.res_type != existing_vg.res.res_type:
                vg_diff.append (
                    VGChange (
                        is_new=False, 
                        is_removed=False, 
                        res_changed=(existing_vg.res.res_type, new_vg.res.res_type),
                        vg=new_vg
                    )
                )
            # ok, nothing changed
            else:
                vg_diff.append (
                    VGChange (
                        is_new=False, 
                        is_removed=False, 
                        res_changed=None,
                        vg=new_vg
                    )
                )
    
    for old_vg in old_vgs:
        if old_vg.res is None: continue
        
        if find_a_match(old_vg, new_vgs) is None:
            vg_diff.append(
                VGChange(
                    is_new=False, 
                    is_removed=True, 
                    res_changed=(old_vg.res.res_type, ""),
                    vg=old_vg
                )
            )

    return vg_diff

class SketchState(BaseModel):
    """
    SketchState contains the current version of 
    """
    state_id : int # State ID
    change : SketchChange # Change that caused this state
    iml_code : str # Actual IML code
    status : FormalizationStatus # Formalization status 
    error : Optional[str] = None # Error feedback from ImandraX
    
    vgs : list[VG] # List of current verification goals
    old_vgs : list[VG] # List of VGs in the last (good) state with admissable model
    decomps : list[RegionDecomp] = [] # Dictionary of decomposition requests/results

    vg_changes : list[VGChange] = Field (
        default_factory=lambda data:
            calc_vg_difference(data['vgs'], data['old_vgs']) if 'vgs' in data and 'old_vgs' in data else []
      )  # Changes in VGs
    
    # Summary of VG changes
    summary : ChangeSummary = Field (
        default_factory=lambda data:
            ChangeSummary(vg_changes=data['vg_changes'])
    )

class Sketch(BaseModel):
    """
    Sketch represent versions of a model
    """

    sketch_id : str = Field(default_factory=lambda: str(uuid.uuid4())[:8])
    processing : bool = False # TODO do we need this for the UI?
    anchor_model_path : str # relative path of the model that was used as the anchor 
    init_iml_model : str = "" # the initial model
    sketch_states : list[SketchState] = [] # here we keep track of changes in the state
    base_models : Dict[str, Model] # Dict (by relative path) of base models that went into this

    def get_state_by_idx(self, idx:int) -> SketchState:
        """
        Return SketchState for specific state ID
        """

        if not (0 <= idx and idx < len(self.sketch_states)):
            raise ValueError(f"Out of bounds state id: {idx}")
        
        return self.sketch_states[idx]

    def rollback(self, target_state_id:int|None=None):
        """
        rollback states 
        """

        if target_state_id is None:
            # We should just be able to 
            if len(self.sketch_states) in [0,1]:
                raise Exception (f"Not enough states[={len(self.sketch_states)}] to rollback!")
            self.sketch_states = self.sketch_states[1:]
        
        else:
            if not (0 <= target_state_id and target_state_id < len(self.sketch_states)):
                raise Exception (f"Invalid target state id[={target_state_id}] specified!")
            
            self.sketch_states = self.sketch_states[:(target_state_id+1)]


        
    def state_ids(self) -> list[int]:
        """
        Return list of SketchState IDs
        """
        if self.sketch_states:
            return list(range(len(self.sketch_states)))
        return []

    def latest_state(self) -> SketchState|None:
        """
        Return the latest sketch state if it's available
        """

        if len(self.sketch_states):
            return self.sketch_states[0]
        
        return None
    
    def last_good_state(self) -> SketchState|None:
        """
        Return last sketch state that has an admissable model
        """

        for i in range(len(self.sketch_states)-1, -1, -1):
            if self.sketch_states[i].status == FormalizationStatus.TRANSPARENT:
                return self.sketch_states[i]
        
        return None

    def gen_init_task(self) -> SketchChangeTask:
        """
        This is the task that needs to be run to populate the initial state
        """

        return SketchChangeTask (
            sketch_id = self.sketch_id,
            iml_code = self.init_iml_model,
            change = SketchChgSetModel(new_iml_code=self.init_iml_model)
        )

    def get_task (self, change: SketchChange) -> SketchChangeTask:
        """
        Return a SketchChangeTask that can be sent to the worker to be analyzed by ImandraX
        """
        
        # Writing it this way so it's easier for pyright to pick up
        state = self.latest_state()

        if state is None:
            raise Exception(f"Attempting to create a task but we have no initial state!")
                
        new_iml_model = self.process_change(state.iml_code, change)

        # Now we have a sketch change task!
        return SketchChangeTask(
            sketch_id=self.sketch_id,
            change = change,
            iml_code = new_iml_model
        )

    def apply_change_result (self, change_result : SketchChangeResult) -> ChangeSummary:
        """ 
        Apply the result of running change task to the state and get the ChangeSummary as the result
        """

        if len(self.sketch_states) == 0 or self.last_good_state() is None:
            state_id = 0
            old_vgs = []
            self.init_iml_model = change_result.task.iml_code
        else:
            state_id = len(self.sketch_states)+1
            old_vgs = self.last_good_state().vgs # pyright: ignore
            
        new_state = SketchState (
            state_id = state_id,
            change = change_result.task.change,
            iml_code = change_result.task.iml_code,

            status = FormalizationStatus.TRANSPARENT if change_result.success else FormalizationStatus.INADMISSIBLE,

            error = change_result.error,
            vgs = change_result.vgs,
            old_vgs= old_vgs,
            decomps = change_result.decomps
        )

        self.sketch_states.insert(0, new_state)

        return new_state.summary

    def get_model_changes(self) -> Dict[str, str]:
        """
        Generate changes to the underlying models
        """

        return {}

    def process_change(self, iml_code:str, change:SketchChange) -> str:
        """
        Process the change command for the specified code and return its updated version
        """

        if isinstance(change, SketchChgSetModel):
            updated_iml_code = change.new_iml_code

        elif isinstance(change, SketchChgInsertDef):
            try:
                updated_iml_code = add_definition(
                    source = iml_code, 
                    new_def = change.new_def_code
                )
            except Exception as e:
                raise Exception (f"Failed to insert: {e}")
        
        elif isinstance(change, SketchChgModifyDef):
            try:
                if isinstance (change.def_name_or_vg_idx, int):
                    updated_iml_code = update_definition(
                        source=iml_code, name="",
                        new_def=change.new_def_body,
                        verify_index=change.def_name_or_vg_idx
                    )
                else:
                    updated_iml_code = update_definition(
                        source=iml_code, 
                        name=change.def_name_or_vg_idx, 
                        new_def=change.new_def_body
                    )
            except Exception as e:
                raise Exception (f"Failed to update definition: {e}")

        elif isinstance (change, SketchChgDeleteDef):
            try:
                if isinstance(change.def_name_or_vg_idx, int):
                    updated_iml_code = remove_definition(
                        source=iml_code, name="", verify_index=change.def_name_or_vg_idx
                    )
                else:
                    updated_iml_code = remove_definition(
                        source=iml_code, name=change.def_name_or_vg_idx
                    )
            except Exception as e:
                raise Exception (f"Failed to delete definition: {e}")

        else:
            raise Exception (f"Unsupported change event: {change}")
        
        return updated_iml_code
  
    def __repr__(self):
        return f"Sketch [ID = {self.sketch_id}; Anchor model path = {self.anchor_model_path}]"

class SketchContainer(BaseModel):
    """
    Container for helping us keep track of the sketches
    """

    sketches : Dict[str, Sketch] = {}

    def add (self, new_sketch:Sketch) -> None:
        """
        Add spetch to the container
        """
        if new_sketch.sketch_id in self.sketches:
            raise ValueError(f"Sketch with specified ID is already present: {new_sketch.sketch_id}")

        self.sketches[new_sketch.sketch_id] = new_sketch

    def from_id(self, sketch_id) -> Sketch|None:
        """
        Return sketch from sketch ID, None if it's missing
        """

        if sketch_id in self.sketches:
            return self.sketches[sketch_id]
        return None

    def ids(self) -> list[str]:
        """ 
        List of available sketch IDs
        """
        return list(self.sketches.keys())

    def list_sketches(self) -> list[Dict[str,str]]:
        """ 
        List available sketches
        """
        
        res = []
        for sketch_id in self.sketches:
            res.append ({
                'sketch_id': sketch_id,
                'anchor_model': self.sketches[sketch_id].anchor_model_path
            })
        
        return res

    def toJSON(self):
        """
        Return JSON dict
        """
        return self.model_dump_json()

    @staticmethod
    def fromJSON(j : str|dict):
        """
        Create a SketchesContainer object from json dict
        """

        if isinstance (j, str):
            return SketchContainer.model_validate_json(j)
        else:
            return SketchContainer.model_validate(j)
  
    def __rich__ (self):
        """ Return the list of rich representations of the sketches we hold """

        pretty = Pretty(self.toJSON(), indent_guides=True)
        return Panel(pretty, title='SketchContainer', border_style="green")




#
#   Imandra Inc.
#
#   model_task.py
#

from imandra.u.agents.code_logician.command import (
    InitStateCommand,
    InjectFormalizationContextCommand,
    EditStateElementCommand,
    GenFormalizationDataCommand,
    GenModelCommand,
    AgentFormalizerCommand,
    GenVgsCommand,
    GenRegionDecompsCommand,
    GenTestCasesCommand,
    EmbedCommand,
    SetModelCommand,
    Command
)

from imandra.u.agents.code_logician.base import FormalizationDependency
from imandra.u.agents.code_logician.graph import GraphState
import uuid, datetime
from typing import Optional
from pydantic import BaseModel, Field, model_validator
from enum import StrEnum

class ModelTaskMode(StrEnum):
    """
    Strategy mode
    """
    AGENT = 'Agent'
    MANUAL = 'Manual'

class FormalizationNeed(StrEnum):
    """
    Reasons to perform formalization
    """
    NO_AGENT_STATE = "No_agent_state" # This is the first time we're formalizing the model
    SRC_CODE_CHANGED = "Source_code_changed" # Source code has changed
    CONTEXT_ADDED = "Context_added" # Human feedback has been provided
    DEPS_CHANGED = "Dependencies_changed" # Dependencies (at least one) have changed

class ModelTask(BaseModel):
    """ 
    Model task contains a single task for Code Logician to execute
    """

    rel_path        : str # path of the source file (relative to the source code directory)
    src_code        : str = "" # source code
    context         : Optional[str] = "" # IML source code for dependent models
    dependencies    : list[FormalizationDependency] = [] # IML source code for dependent models
    graph_state     : Optional[GraphState] = None # previous graph state, if available
    language        : str = "Python" # programming language 
    mode            : ModelTaskMode = ModelTaskMode.AGENT # mode used with CodeLogician (either "simple" or "agent")
    gen_vgs         : bool = True # Should we generate verification goals?
    gen_decomps     : bool = True # Should we generate decompositions?
    gen_embeddings  : bool = True # Should we generate embeddings?

    # This is used to submit user-specified commands for specific models
    # if this is set, then we disregard everything else and just return these
    # along with the existing state
    specified_commands : Optional[list[Command]] = None

    # Each task has a unique ID that we'll then use to assign back the result
    # of CL work for this task - if the model's task ID doesn't match the result
    # then we discard it - this way we always keep the most recent result
    task_id : str = Field(default_factory=lambda: str(uuid.uuid4())) # assigned task ID, new one created if not provided
  
    # Same with the timestamp, if it's provided (mostly during de/serialization, 
    # then let's use it otherwise, let's create a new one)
    created_at : datetime.datetime = Field(default_factory=datetime.datetime.now) # Each instance gets a new timestamp

    def commands(self):
        """
        Return a list of CodeLogician agent commands for this task
        """

        if self.specified_commands:
            # We need to return those and add nothing else
            return self.specified_commands

        commands : list[Command] = [
            InitStateCommand(src_code=self.src_code, src_lang=self.language)
        ]

        if self.context is not None:
            commands.append(InjectFormalizationContextCommand(context=self.context))
    
        if self.dependencies is not None:
            commands.append(EditStateElementCommand(update={"dependency": self.dependencies}))

        if self.mode == ModelTaskMode.MANUAL:
            commands.append(GenFormalizationDataCommand())
            commands.append(GenModelCommand())

        elif self.mode == ModelTaskMode.AGENT:
            commands.append(
                AgentFormalizerCommand(
                    no_gen_model_hitl=True,
                    max_tries_wo_hitl=3,
                    max_tries=3,
                    no_check_formalization_hitl=True,
                    no_refactor=False
                )
            )
        else:
            raise Exception(f"Attempting unrecognized mode: {self.mode}")

        if self.gen_vgs:
            commands.append(GenVgsCommand(description=""))

        if self.gen_decomps:
            commands.append(GenRegionDecompsCommand(function_name=None))

        #if self.genDecomps or len(self.tests) > 0:    
        #for i in self.tests:
        #    commands.append(GenTestCasesCommand(decomp_idx=i))

        if self.gen_embeddings:
            commands.append(EmbedCommand())

        return commands

    def toJSON(self):
        """
        Convert to a JSON
        """
        return self.model_dump_json()
  
    @staticmethod
    def fromJSON(j : dict|str):
        """
        fromJSON
        """
        if isinstance(j, str):
            return ModelTask.model_validate_json(j)
        else:
            return ModelTask.model_validate(j)

    def __repr__(self):
        """ """
        return f"Base ModelTask with ID = {self.task_id}; path={self.rel_path}"

class UserManualIMLEditTask(ModelTask, BaseModel):
    """
    This updates the graph state from a manual user IML edit
    """

    iml_code : str # user provided IML code

    def commands(self):
        """
        We're just going to invoke the commands 
        """

        if self.graph_state:
            commands = [
                SetModelCommand(model=self.iml_code)
            ]
        else:
            commands : list[Command] = [
                InitStateCommand(src_code=self.src_code, src_lang=self.language),
                SetModelCommand(model=self.iml_code)

            ]

        return commands




from pathlib import PurePath
from collections import defaultdict

import httpx, json

from strategy.state import StrategyState

endpoint = "http://127.0.0.1:8000/strategy/states"
data = httpx.get(endpoint)

j = json.loads(data.text.strip("'"))

strat_states = {}
for path in j:
    strat_state = StrategyState.fromJSON(j[path])
    strat_states[path] = strat_state

print (strat_state)


#
#   Imandra Inc.
#
#   help.py
#   

from textual.screen import Screen
from textual.widgets import Footer, MarkdownViewer
from ..common import MyHeader, text_read, local_file

class HelpScreen(Screen):
    def __init__(self):
        Screen.__init__(self)
        self.title = "Help"

    def compose(self):
        yield MyHeader()
        text = text_read(local_file("data/help.md"))
        markdown_viewer = MarkdownViewer(text, show_table_of_contents=True)
        markdown_viewer.code_indent_guides = False
        yield markdown_viewer
        yield Footer()